"""
French for Vellum

> Vellum is prepared animal skin or "membrane", typically used as a material for
> writing on. Parchment is another term > for this material, and if vellum is
> distinguished from this, it is by vellum being made from calfskin, as opposed to
> that from other animals,[1] or otherwise being of higher quality

Tools to automatically reformat docstrings based using numpydoc format.

"""

import sys
import textwrap
from pathlib import Path

from .ref import NumpyDocString, main

__version__ = "0.0.6"


def parse(input):
    """
    parse an input string into token/tree.

    For now only return a list of tokens

    """
    tokens = []
    for l in input.splitlines():
        tokens.extend(l.split(" "))
    return tokens


def transform(tokens):
    """
    Accumulate tokens in lines.

    Add token (and white spaces) to a line until it overflow 80 chars.
    """
    lines = []
    current_line = []
    for t in tokens:
        if sum([len(x) + 1 for x in current_line]) + len(t) > 80:
            lines.append(current_line)
            current_line = []
        current_line.append(t)
    if current_line:
        lines.append(current_line)
    return lines


def format(lines):
    s = ""
    for l in lines:
        s = s + " ".join(l)
        s = s + "\n"
    return s[:-1]


def compute_indents(lines):
    """
    Given a list of lines, compute indentation in number of spaces.

    Indentation is only supported if spaces, tabs raise a NotImplementedError as we don't know how wide a tab is.
    We also treat complete blank lines as `None` indentation.

    """
    assert isinstance(lines, list)
    results = []
    for l in lines:
        s = l.lstrip()
        if not s:
            results.append(None)
            continue
        indent = len(l) - len(s)
        if "\t" in l[:indent]:
            raise NotImplementedError

        results.append(indent)
    return results


class TryNext(Exception):
    pass


class Header:
    def __init__(self, title, level):
        self.title = title
        # if str(title) not in NumpyDocString.sections.keys():
        #    print(f'??? |{title}|')
        self.level = level

    def __repr__(self):
        return f"<Header {self.level}:  {self.title} >"

    def __str__(self):
        tt = str(self.title)
        return tt + "\n" + "=" * len(tt)

    def _repr_html_(self):
        return f"<h{self.level}>" + str(self.title) + f"</h{self.level}>"

    @classmethod
    def parse(cls, lines):
        if len(lines) < 2:
            raise TryNext
        assert lines
        warnings = []
        title, _, wn = Raw.parse(lines)
        lgth = len(title.lines[0])
        if (
            len(set(lines[1])) == 1
            and len(lines[1]) != 1
            and len(lines[1]) != lgth
            and ">>>" not in lines[0]
            and "::" not in lines[1]
        ):

            warnings.append("======= WRONG LEN? ======")
            warnings.append("L0: " + lines[0])
            warnings.append("L1: " + lines[1])
            warnings.append("=========================")
        level = allunders(lines[1], lgth)
        return cls(title, level), lines[2:], wn


def allunders(line, lenght):

    if not len(set(line)) == 1 or not len(line) == lenght:
        raise TryNext
    if next(iter(set(line))) not in "-=~`":
        raise TryNext
    return 1


class Any:
    def __init__(self, line):
        assert isinstance(line, str)
        self.lines = [line]

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.lines!r} >"

    def __str__(self):
        return self.lines[0]

    @classmethod
    def parse(cls, lines):
        return cls(lines[0]), lines[1:], []


class Raw(Any):
    pass


class RawTilNextHeader:
    @classmethod
    def parse(cls, lines):
        ll = []
        for i, l in enumerate(lines):
            try:
                Header.parse(lines[i:])
                break
            except TryNext:
                pass

            ll.append(l)
        else:
            return cls(ll), [], []

        return cls(ll), lines[i:], []

    def __init__(self, items):
        assert isinstance(items, list)
        self.items = items

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.items!r} >"

    def __str__(self):
        return "\n".join(self.items)

    def _repr_html_(self):
        return """<pre>""" + str(self) + """</pre>"""


class DescriptionList:
    @classmethod
    def parse(cls, lines):
        dct = {}
        key, values = None, []
        for i, l in enumerate(lines):
            try:
                Header.parse(lines[i:])
                dct[key] = values
                break
            except TryNext:
                pass

            if not l.startswith(" "):
                dct[key] = values
                key, values = l.strip(), []
                if "," in key:
                    raise TryNext
            else:
                values.append(l)
        if None in dct:
            del dct[None]
        return cls(dct), lines[i:], []

    def __init__(self, items):
        assert isinstance(items, dict)
        self.items = items

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.items!r} >"

    def __str__(self):
        def _f(v):
            return "\n".join(v)

        return "\n".join(f"{k}:\n{_f(v)}" for k, v in self.items.items())

    def _repr_html_(self):
        def _f(v):
            return "\n".join(v)

        return (
            """<dl>"""
            + "\n".join(
                f"<dt>{k}</dt>\n<dd>{_f(v)}</dd>" for k, v in self.items.items()
            )
            + """</dl>"""
        )


class Listing:
    @classmethod
    def parse(cls, lines):
        assert "," in lines[0], lines[0]
        listing = [l.strip() for l in lines[0].split(",")]
        return cls(listing), lines[1:], []

    def __init__(self, listing):
        self.listing = listing

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.listing!r} >"

    def __str__(self):
        return ",".join([f"{k}" for k in self.listing])

    def _repr_html_(self):
        return (
            """<ul>""" + "\n".join(f"<li>{k}</li>" for k in self.listing) + """</ul>"""
        )


class Base:
    def attrs(self):
        stuff = {}
        for it in dir(self):
            if it.startswith("_"):
                continue
            att = getattr(self, it)
            if callable(att):
                continue
            stuff[it] = att
        return stuff

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} "
            + ", ".join([k + ":" + repr(x) for k, x in self.attrs().items()])
            + ">"
        )


class EntryParser(Base):
    @classmethod
    def parse(cls, lines):
        if len(lines) == 1:
            l0, l1 = lines[0], ""
            rest = []
        elif len(lines) < 3:
            l0, l1 = lines
            rest = []
        else:
            l0, l1, *rest = lines
        indent = len(l1) - len(l1.lstrip())
        i = 0
        if indent:
            cont = [l1]
            for i, l in enumerate(rest):
                if l.startswith(" " * indent) or not l.strip():
                    cont.append(l)
                else:
                    break
            else:
                i += 1
        else:
            i = 0
            cont = []
            rest = lines[1:]
        if ":" in l0:
            try:
                head, t = [x.strip() for x in l0.split(":", maxsplit=1)]
            except ValueError:
                print("... Entry TryNext", lines[:5])
                raise TryNext
        else:
            head, t = l0.strip(), ""

        if " " in head:
            if not t and not cont:
                # print('... list of things ? ', head)
                return [cls(h.strip(), "", []) for h in head.split(",")], rest[i:]

            if "See Also" in head:
                print("-------------->", lines[:3])

        return [cls(head, t, cont)], rest[i:]

    def __init__(self, head, t, rest):
        # assert (head.strip() or t.strip() or [r.strip() for r in rest])
        self.head = head
        self.t = t
        self.rest = rest

    def __str__(self):
        r = "\n".join(self.rest)
        if r:
            extra = f"\n{r}"
        else:
            extra = ""
        return f"""{self.head}:{self.t}{extra}"""

    def _format_head(self, head, resolver):
        return resolver(head)

    def _format_type(self, key, resolver):
        if self.rest:
            return key
        else:
            return ""

    def _format_core(self, core, resolver):
        if self.rest:
            return "<pre>" + "\n".join(core) + "</pre>"
        else:
            return f"<pre>{self.t}</pre>"

    def _repr_html_(self, resolver):

        return f"<dt>{self._format_head(self.head, resolver)}:{self._format_type(self.t, resolver)}</dt><dd>{self._format_core(self.rest, resolver)}</dd>"


class DeflistParser(Base):
    def __init__(self, entries):
        self.entries = entries

    @classmethod
    def parse(cls, lines):
        ents = []
        while lines:
            if not lines[0].strip():
                lines = lines[1:]
                continue
            try:
                Header.parse(lines)
                break
            except TryNext:
                pass
            e, lines = EntryParser.parse(lines)
            ents.extend(e)

        return cls(ents), lines

    def __str__(self):
        return "\n".join(str(x) for x in self.entries)

    def _repr_html_(self, resolver=None):

        return (
            """<dl> (Deflist)"""
            + "\n".join(x._repr_html_(resolver) for x in self.entries)
            + """</dl>"""
        )


class Mapping:
    @classmethod
    def parse(cls, lines):
        mapping = {}
        k = None
        for i, l in enumerate(lines):
            if not l.strip():
                continue
            try:
                Header.parse(lines[i:])
                break
            except TryNext:
                pass
            if l.startswith(" ") and k:
                try:
                    mapping[k.strip()] += l.strip()
                except TypeError:
                    raise TryNext

            if ":" in l:
                k, v = l.split(":", maxsplit=1)
                mapping[k.strip()] = v
            elif "," not in l:
                k, v = l.strip(), None
                mapping[k] = v
            else:
                for k in l.split(","):
                    mapping[k.strip()] = None

        return cls(mapping), lines[i:], []

    def __init__(self, mapping):
        self.mapping = mapping

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.mapping!r} >"

    def __str__(self):
        return "\n".join([f"{k}: {v}" for k, v in self.mapping.items()])

    def _format_one(self, key, resolver):
        return resolver(key)

    def _format_pair(self, key, value, resolver):
        k = self._format_one(key, resolver)
        v = value
        return f"<dt>{k}</dt>\n<dd>{v}</dd>"

    def _repr_html_(self, resolver):

        return (
            """<dl> (Mapping)"""
            + "\n".join(
                [self._format_pair(k, v, resolver) for k, v in self.mapping.items()]
            )
            + """</dl>"""
        )


class CodeBlock:
    @classmethod
    def parse(cls, lines):
        if not lines[0].startswith((">>>", "    >>>")):
            raise TryNext

        _lines = []
        for i, l in enumerate(lines):
            if not l.strip():
                break
            _lines.append(l)
        else:
            return cls(_lines), [], []
        return cls(_lines), lines[i:], []

    def __init__(self, lines):
        self.lines = lines

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.lines!r} >"

    def _repr_html_(self):
        return "<pre>" + "\n".join(self.lines) + "</pre>"

    def __str__(self):
        return "\n".join(self.lines)


class Doc:
    @classmethod
    def parse(cls, lines, *, name=None, sig=None):
        parsed = []
        cl = len(lines)
        warnings = []
        while lines:
            for t in (
                Section.parse,
                Header.parse,
                BlankLine.parse,
                CodeBlock.parse,
                Paragraph.parse,
                RawTilNextHeader.parse,
                failed,
            ):
                try:
                    node, lines_, wn = t(lines)
                    warnings.extend(wn)
                    if isinstance(node, list):
                        parsed.extend(node)
                    else:
                        parsed.append(node)
                    if len(lines_) >= len(lines):
                        raise ValueError("Could not parse", lines)
                    lines = lines_
                    break
                except TryNext:
                    pass
        return cls(parsed, name=name, sig=sig), warnings

    def __init__(self, nodes, name=None, sig=None):
        self.nodes = nodes
        self.name = name
        self.sig = sig
        self.backrefs = []

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}\n"
            + "\n".join([textwrap.indent(repr(n), "    ") for n in self.nodes])
            + "\n>"
        )

    def see_also(self):
        for i, p in enumerate(self.nodes):
            if isinstance(p, Section) and p.header.title.lines[0] == "See Also":
                break
        else:
            return []

        node = self.nodes[i + 1]
        if isinstance(node, Mapping):
            return node.mapping.keys()
        elif isinstance(node, DeflistParser):
            return [x.head for x in node.entries]
        else:
            print("not a mapping", repr(node))
            pass

    def _repr_html_(self, resolver=lambda x: None):
        base = """<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Document</title>
    <link rel="stylesheet" href="https://fonts.xz.style/serve/inter.css">
    <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/@exampledev/new.css@1.1.2/new.min.css">
</head>
<body>
    {}
    {}
    {}
</body>
</html>
"""

        h1 = ""
        if self.name:
            h1 += self.name
        if self.sig:
            h1 += self.sig
        if h1:
            h1 = "<h1>" + h1 + "</h1>"

        def f_(it):
            return f"<a href=./{it}.html>{it}</a>"

        if self.backrefs:
            backref_repr = ",".join(self.backrefs)
            br_html = "<h1>Back references</h1>" + ", ".join(
                f_(b) for b in self.backrefs
            )
        else:
            br_html = ""

        def _resolver(k):
            ref = resolver(k)
            if ref is None:
                # print("could not resolve", k, f"({self.name})")
                return k
            else:
                # print("resolved", k, f"({self.name})")
                return f"<a href='{ref}.html'>{k}</a>"

        hrepr = []
        for n in self.nodes:
            if isinstance(n, (Mapping, DeflistParser)):
                hrepr.append(n._repr_html_(_resolver))
            else:
                hrepr.append(n._repr_html_())

        return base.format(h1, "\n".join(hrepr), br_html)


class Section:
    """
    start with a header, but have custom parsing because we know about it in numpydoc.
    """

    @classmethod
    def parse(cls, lines):
        warnings = []
        header, rest, wn = Header.parse(lines)
        warnings.extend(wn)
        aliases = {
            "Return": "Returns",
            "Arguments": "Parameters",
            "arguments": "Parameters",
            "additional keyword arguments:": "Parameters",
            "Other parameters": "Other Parameters",
            "Exceptions": "Raises",
        }
        ht = header.title.lines[0]
        if ht in aliases:
            warnings.append(f"Found `{ht}`, did you mean `{aliases[ht]}` ?")
            header.title.lines[0] = aliases[ht]
        ht = header.title.lines[0]
        if ht in (
            "Parameters",
            "Returns",
            "Class Attributes",
            "Options",
            "Attributes",
            "Yields",
            "Raises",
            "Exceptions",
            "Methods",
            "Warns",
            "Other Parameters",
            "Warnings",
            "Arguments",
        ):
            try:
                try:
                    core, rest, _ = Paragraph.parse(rest)
                except TryNext:
                    core, rest = DeflistParser.parse(rest)
            except TryNext:
                core, rest, wn = DescriptionList.parse(rest)
                warnings.extend(wn)
            return [cls(header), core], rest, warnings
        elif header.title.lines[0] in ("See Also", "Returns", "See also"):
            if header.title.lines[0] == "See also":
                header.title.lines[0] = "See Also"
            try:
                core, rest = DeflistParser.parse(rest)
            except TryNext:
                print("Deflist failed trying Mapping... ")
                core, rest, wn = Mapping.parse(rest)
                warnings.extend(wn)
                # core, rest, wn = DescriptionList.parse(rest)
                # warnings.extend(wn)

            return [cls(header), core], rest, warnings
        elif header.title.lines[0] in ("Examples",):
            return [cls(header)], rest, warnings
        elif header.title.lines[0] in ("Notes", "References"):
            return [cls(header)], rest, warnings

        else:
            raise ValueError(repr(header.title.lines[0]))

    def __init__(self, header):
        self.header = header

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.header!r}>"

    def __str__(self):
        return str(self.header)

    def _repr_html_(self):
        return self.header._repr_html_()


class Paragraph:
    @classmethod
    def parse(cls, lines):
        _lines = []
        l0 = lines[0]
        if not l0 or l0.startswith(" "):
            raise TryNext
        if len(lines) >= 2:
            if lines[1].startswith(" ") and lines[1].strip():
                # second line indented this _is_ a deflist
                raise TryNext
        for i, l in enumerate(lines):
            if l and not l.startswith(" "):
                _lines.append(l)
            else:
                break
        if not _lines:
            raise TryNext
        return cls(_lines), lines[i + 1 :], []

    def __init__(self, lines):
        self.lines = lines

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.lines}>"

    def __str__(self):
        return "\n".join(self.lines)

    def _repr_html_(self):
        return "<p>" + " ".join(self.lines) + "</p>"


class BlankLine:
    @classmethod
    def parse(self, lines):
        if not lines[0].strip():
            return BlankLine(), lines[1:], []
        raise TryNext

    def __init__(self):
        pass

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def __str__(self):
        return ""

    def _repr_html_(self):
        return ""


def failed(lines):
    raise ValueError("nothign managed to parse", lines)


def parsedoc(doc, *, name=None, sig=None):

    try:
        NumpyDocString(doc)
        lines = dedentfirst(doc).splitlines()
        return Doc.parse(lines, name=name, sig=sig)
    except Exception:
        return ""


def find_indent_blocks(lines):
    """
    Given a list of lines find _block_ by level of indentation

    - A block is considered a sequence of one or more lines, separated
    - once one level of indentation is encounter we don't split anymore, it will be up to the sub-parser.
    - A whitespace or empty line yield another block.
    """
    if isinstance(lines, str):
        raise ValueError("split please")
    indents = compute_indents(lines)
    assert len(indents) == len(lines)
    l0 = lines[0]
    indent_level = indents[0]
    if indent_level is None:
        indent_level = 0
    assert indent_level == 0
    current_block = [l0]

    n_empty_lines = 0

    blocks = []
    for new_level, l in zip(indents[1:], lines[1:]):
        if new_level is None:
            n_empty_lines += 1
            continue

        if n_empty_lines and new_level == 0:
            current_block.extend([""] * n_empty_lines)
            n_empty_lines = 0
            blocks.append((indent_level, current_block))
            current_block = [l]
            indent_level = new_level
            continue
        if n_empty_lines:
            current_block.extend([""] * n_empty_lines)
            n_empty_lines = 0

        if indent_level == 0 and new_level:
            # start a new block
            blocks.append((0, current_block))
            current_block = [l[new_level:]]
            indent_level = new_level
            continue

        # we are already in indented blocks.
        if new_level >= indent_level:
            current_block.append(l[indent_level:])
        elif new_level < indent_level:
            blocks.append((indent_level, current_block))
            current_block = [l]
            indent_level = new_level
            continue
    if current_block:
        blocks.append((indent_level, current_block))
        current_block = []

    assert len(current_block) == 0
    return blocks


def dedentfirst(docstring):
    from textwrap import dedent

    lines = docstring.splitlines()
    ln = dedent("\n".join(lines[1:])).splitlines()
    l0 = lines[0]
    return "\n".join([dedent(l0)] + ln)


def reformat(input):
    return format(transform(parse(format(transform(parse(input))))))


# def main():
#
#    for filename in sys.argv[1:]:
#        p = Path(filename)
#
#        with p.open() as f:
#            data = f.read()
#            print(reformat)


if __name__ == "__main__":
    main()



import glob

import pytest
from velin import compute_indents, find_indent_blocks, reformat

test_files = glob.glob("examples/*.rst")

expected = [f[:-3] + "expected" for f in test_files]


@pytest.mark.parametrize("test_input,expected", zip(test_files, expected))
def test_reformat_1(test_input, expected):
    with open(test_input) as f:
        inp = f.read()
    with open(expected) as f:
        exp = f.read()
    assert reformat(inp) == exp.strip("\n")


@pytest.mark.parametrize(
    "test_input,expected",
    [
        (
            """this
is an
example
""",
            [0, 0, 0],
        ),
        (
            """this
  
example
""",
            [0, None, 0],
        ),
        (
            """ this
  
example
""",
            [1, None, 0],
        ),
        (
            """ this
  
  example
""",
            [1, None, 2],
        ),
    ],
)
def test_blocks(test_input, expected):
    assert compute_indents(test_input.splitlines()) == expected



import argparse
import ast
import difflib
import json
import re
import sys
from configparser import ConfigParser
from pathlib import Path
from textwrap import indent

import numpydoc.docscrape as nds
from numpydoc.docscrape import Parameter

try:
    from there import print
except ImportError:
    pass

from .examples_section_utils import reformat_example_lines


class NodeVisitor(ast.NodeVisitor):
    def __init__(self):
        self.items = []

    def visit_FunctionDef(self, node):
        # we can collect function args, and _could_ check the Parameter section.
        # for nn in node.args.args:
        #    nnn = nn.arg
        # for k in [l for l in dir(node) if not l.startswith('_')]:
        #    sub = getattr(node, k)
        #    print(k, sub)
        #    for u in [l for l in dir(sub) if not l.startswith('_')]:
        #        ss = getattr(sub, u)
        #        print('    ', u, ss)

        # node.args.kwarg # single object
        # node.args.vararg # object
        # node.args.ks_defaults #list
        # print(node.returns)
        # node.args.defaults +

        meta = [
            arg.arg
            for arg in node.args.posonlyargs + node.args.args + node.args.kwonlyargs
        ]
        #    print(arg.arg)
        #    for k in [l for l in dir(arg) if not l.startswith('_')]:
        #        sub = getattr(arg, k)
        #        print('    ', k, sub)

        self.items.append((node, meta))
        self.generic_visit(node)

    def visit_ClassDef(self, node):
        # we can collect function args, and _could_ check the Parameter section.
        # for nn in node.args.args:
        #    nnn = nn.arg

        # self.items.append((node, None))
        self.generic_visit(node)


BLACK_REFORMAT = True


class NumpyDocString(nds.NumpyDocString):
    """
    subclass a littel bit more lenient on parsing
    """

    aliases = {
        "Parameters": (
            "options",
            "parameter",
            "parameters",
            "paramters",
            "parmeters",
            "paramerters",
            "arguments",
        ),
        "Attributes": ("properties",),
        "Yields": ("signals",),
    }

    def normalize(self):
        """
        Apply a bunch of heuristic that try to normalise the data.
        """
        if (params := self["Parameters"]) :
            for i, p in enumerate(params):
                if not p.type and ":" in p.name:
                    if p.name.startswith(".."):
                        continue
                    name, type_ = [_.strip() for _ in p.name.split(":", maxsplit=1)]
                    params[i] = nds.Parameter(name, type_, p[2])

    def parse_examples(self, lines, indent=4):
        # this is bad practice be we do want normalisation here for now
        # to  check that parse->format->parse is idempotent.
        # this can be done if we had a separate "normalize" step.
        global BLACK_REFORMAT
        if BLACK_REFORMAT:
            try:
                lines = reformat_example_lines(lines, indent=indent)
            except Exception:
                print("black failed")
                print("\n".join(lines))
                raise

        return lines

    def to_json(self):

        res = {k: v for (k, v) in self.__dict__.items() if ((k not in {"_doc"}) and v)}
        res["_parsed_data"] = {k: v for (k, v) in res["_parsed_data"].items() if v}

        return res

    @classmethod
    def from_json(cls, obj):
        nds = cls("")
        nds.__dict__.update(obj)
        # print(obj['_parsed_data'].keys())
        nds._parsed_data["Parameters"] = [
            Parameter(a, b, c) for (a, b, c) in nds._parsed_data.get("Parameters", [])
        ]

        for it in (
            "Returns",
            "Yields",
            "Extended Summary",
            "Receives",
            "Other Parameters",
            "Raises",
            "Warns",
            "Warnings",
            "See Also",
            "Notes",
            "References",
            "Examples",
            "Attributes",
            "Methods",
        ):
            if it not in nds._parsed_data:
                nds._parsed_data[it] = []
        for it in ("index",):
            if it not in nds._parsed_data:
                nds._parsed_data[it] = {}
        return nds

    def __init__(self, *args, **kwargs):
        self.ordered_sections = []
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        if key in ["Extended Summary", "Summary"]:
            value = [d.rstrip() for d in value]

        if key in ("Examples"):
            value = self.parse_examples(value)
        super().__setitem__(key, value)
        assert key not in self.ordered_sections
        self.ordered_sections.append(key)

    def _guess_header(self, header):
        if header in self.sections:
            return header
        # handle missing trailing `s`, and trailing `:`
        for s in self.sections:
            if s.lower().startswith(header.rstrip(":").lower()):
                return s
        for k, v in self.aliases.items():
            if header.lower() in v:
                return k
        raise ValueError("Cound not find match for section:", header)

    def _read_sections(self):
        for name, data in super()._read_sections():
            name = self._guess_header(name)
            yield name, data

    def _parse_param_list(self, *args, **kwargs):
        """
        Normalize parameters
        """
        parms = super()._parse_param_list(*args, **kwargs)
        out = []
        for name, type_, desc in parms:
            out.append(
                nds.Parameter(name.strip(), type_.strip(), [d.rstrip() for d in desc])
            )
        return out


def w(orig):
    """Util function that shows whitespace as `·`."""
    ll = []
    for l in orig:
        if l[0] in "+-":
            # ll.append(l.replace(' ', '⎵'))
            ll.append(l.replace(" ", "·"))

        else:
            ll.append(l)
    lll = []
    for l in ll:
        if l.endswith("\n"):
            lll.append(l[:-1])
        else:
            lll.append(l[:])
    return lll


class Conf:
    """

    Here are some of the config options

    - F100: Items spacing in Return/Raise/Yield/Parameters/See Also
        - A(auto) - spaces if some sections have blank lines. (default)
        - B(compact) - no spaces between items, ever.
        - C(spaced)  - spaces between items, always

    """

    defaults = {"F100": "A"}

    def __init__(self, conf):
        self._conf = conf

    def __getattr__(self, key):
        k = key[:-1]
        v = key[-1]
        return self._conf.get(k, self.defaults[k]) == v


class SectionFormatter:
    """
    Render section of the docs, based on some configuration. Not having
    configuration is great, but everybody has their pet peevs, and I'm hpping we
    can progressively convince them to adopt a standard.
    """

    def __init__(self, *, conf):
        self.config = Conf(conf)

    @classmethod
    def format_Signature(self, s, compact):
        return s + "\n"

    @classmethod
    def format_Summary(self, s, compact):
        if len(s) == 1 and not s[0].strip():
            return ""
        return "\n".join(s) + "\n"

    @classmethod
    def format_Extended_Summary(self, es, compact):
        return "\n".join(es) + "\n"

    def _format_ps(self, name, ps, compact):
        force_compact = self.config.F100B
        res, try_other = self._format_ps_pref(name, ps, compact=True)
        if (not try_other and self.config.F100A) or self.config.F100B:
            return res
        return self._format_ps_pref(name, ps, compact=False)[0]

    def _format_ps_pref(self, name, ps, *, compact):
        try_other = False
        out = name + "\n"
        out += "-" * len(name) + "\n"
        for i, p in enumerate(ps):
            if (not compact) and i:
                out += "\n"
            if p.type:
                out += f"""{p.name.strip()} : {p.type.strip()}\n"""
            else:
                out += f"""{p.name.strip()}\n"""
            if p.desc:
                if any([l.strip() == "" for l in p.desc]):
                    try_other = True

                out += indent("\n".join(p.desc), "    ")
                out += "\n"
        return out, try_other

    def format_Parameters(self, ps, compact):
        return self._format_ps("Parameters", ps, compact)

    def format_Methods(self, ps, compact):
        return self._format_ps("Methods", ps, compact)

    def format_Other_Parameters(self, ps, compact):
        return self._format_ps("Other Parameters", ps, compact)

    @classmethod
    def format_See_Also(cls, sas, compact):

        res = cls.format_See_Also_impl(sas, True, force_compact=compact)
        if res is not None:
            return res
        return cls.format_See_Also_impl(sas, False, force_compact=compact)

    @classmethod
    def format_See_Also_impl(cls, sas, compact, force_compact):
        out = "See Also\n"
        out += "--------\n"

        for a, b in sas:
            if b:
                desc = b[0]
            else:
                desc = None
            if len(b) > 1:
                rest_desc = b[1:]
            else:
                rest_desc = []
            _first = True
            for ref, type_ in a:
                if not _first:
                    out += ", "
                if type_ is not None:
                    out += f":{type_}:`{ref}`"
                else:
                    out += f"{ref}"
                _first = False

            if desc:
                if len(a) > 1 or (not compact):
                    out += f"\n    {desc}"
                else:
                    attempt = f" : {desc}"
                    if len(out.splitlines()[-1] + attempt) > 75 and not force_compact:
                        return None
                    out += attempt
            for rd in rest_desc:
                out += "\n    " + rd
            out += "\n"
        return out

    @classmethod
    def format_References(cls, lines, compact):
        out = "References\n"
        out += "----------\n"
        out += "\n".join(lines)
        out += "\n"
        return out

    @classmethod
    def format_Notes(cls, lines, compact):
        out = "Notes\n"
        out += "-----\n"
        out += "\n".join(lines)
        out += "\n"
        return out

    @classmethod
    def format_Examples(cls, lines, compact):
        out = "Examples\n"
        out += "--------\n"
        out += "\n".join(lines)
        out += "\n"
        return out

    @classmethod
    def format_Warnings(cls, lines, compact):
        out = "Warnings\n"
        out += "--------\n"
        out += "\n".join(lines)
        out += "\n"
        return out

    @classmethod
    def format_Warns(cls, ps, compact):
        return cls.format_RRY("Warns", ps)

    @classmethod
    def format_Raises(cls, ps, compact):
        return cls.format_RRY("Raises", ps)

    @classmethod
    def format_Yields(cls, ps, compact):
        return cls.format_RRY("Yields", ps)

    @classmethod
    def format_Returns(cls, ps, compact):
        return cls.format_RRY("Returns", ps)

    @classmethod
    def format_Attributes(cls, ps, compact):
        return cls.format_RRY("Attributes", ps)

    @classmethod
    def format_RRY(cls, name, ps):
        out = name + "\n"
        out += "-" * len(name) + "\n"

        if name == "Returns":
            if len(ps) > 1:
                # do heuristic to check we actually have a description list and not a paragraph
                pass

        for i, p in enumerate(ps):
            # if i:
            #    out += "\n"
            if p.name and p.type:
                out += f"""{p.name.strip()} : {p.type.strip()}\n"""
            elif p.name:
                out += f"""{p.name.strip()}\n"""
            else:
                out += f"""{p.type.strip()}\n"""
            if p.desc:
                out += indent("\n".join(p.desc), "    ")
                out += "\n"
        return out


def dedend_docstring(docstring):
    import textwrap

    lines = docstring.splitlines()
    if len(lines) >= 2:
        l0, *lns = docstring.split("\n")
        l0 = textwrap.dedent(l0)
        lns = textwrap.dedent("\n".join(lns)).split("\n")
        docstring = [l0] + lns
    else:
        docstring = textwrap.dedent(docstring).split("\n")
    return "\n".join(docstring)


def compute_new_doc(docstr, fname, *, level, compact, meta, func_name):
    """
    compute a new docstring that shoudl be numpydoc compliant.

    Parameters
    ----------
    docstr : str
        docstring to reformat
    fname : str
        filename of the file beign reformatted (for error messages)
    level : int
        indentation level
    compact : bool
        use compact formating in definition list.
    meta : list
        meta info about the function to verify docstrings, for example
        list of parameters.
    func_name : str
        function name for debug.

    Returns
    -------
    str
        new docstring
    Numpydoc
        parsed numpydoc object

    """
    INDENT = level * " "
    NINDENT = "\n" + INDENT
    original_docstr = docstr
    if len(docstr.splitlines()) <= 1:
        return "", NumpyDocString("")
    shortdoc = bool(docstr.splitlines()[0].strip())
    short_with_space = False
    if not docstr.startswith(NINDENT):
        if original_docstr[0] == " ":
            short_with_space = True

    long_end = True
    long_with_space = True
    if original_docstr.splitlines()[-1].strip():
        long_end = False
    if original_docstr.splitlines()[-2].strip():
        long_with_space = False

    doc = NumpyDocString(dedend_docstring(docstr))
    doc.normalize()
    if (params := doc["Parameters"]) and meta:

        a = [o.strip() for p in params for o in p.name.split(",")]
        if meta[0] in ["self", "cls"]:
            meta = meta[1:]
        doc_missing = set(meta) - set(a) - {"kwargs", "cls"}
        doc_extra = {x for x in set(a) - set(meta) if not x.startswith("*")} - {
            "kwargs",
            "cls",
        }
        if len(doc_missing) == len(doc_extra) == 1:

            print(fname)
            for i, p in enumerate(doc["Parameters"]):
                if p.name in doc_extra:
                    name = p.name
                    new_name = list(doc_missing)[0]
                    doc["Parameters"][i] = nds.Parameter(new_name, *p[1:])
                    print("renamed", name, "to", list(doc_missing)[0])
                    break
            else:
                print("  could not fix:", doc_missing, doc_extra)
        else:
            if doc_missing and doc_extra:
                print(f"{fname}:{func_name}")
                print("  missing:", doc_missing)
                print("  extra:", doc_extra)
            elif doc_missing or doc_extra:
                print(f"{fname}:{func_name}")
                print("  missing:", doc_missing)
                print("  extra:", doc_extra)

    fmt = ""
    start = True
    # ordered_section is a local patch to that records the docstring order.
    if compact:
        df = SectionFormatter(conf={"F100": "B"})
    else:
        df = SectionFormatter(conf={"F100": "A"})
    for s in getattr(doc, "ordered_sections", doc.sections):
        if doc[s]:
            f = getattr(df, "format_" + s.replace(" ", "_"))
            res = f(doc[s], compact)
            if not res:
                continue
            if not start:
                fmt += "\n"
            start = False
            fmt += res
    fmt = indent(fmt, INDENT) + INDENT

    # hack to detect if we have seen a header section.
    if "----" in fmt or True:
        if long_with_space:
            fmt = fmt.rstrip(" ") + NINDENT
        else:
            fmt = fmt.rstrip(" ") + INDENT
    if shortdoc:
        fmt = fmt.lstrip()
        if short_with_space:
            fmt = " " + fmt
    else:
        fmt = "\n" + fmt
    if not long_end:
        fmt = fmt.rstrip()
    assert fmt
    # we can't just do that as See Also and a few other would be sphinxified.
    # return indent(str(doc),'    ')+'\n    '
    return fmt, doc


def reformat_file(data, filename, compact, unsafe, fail=False):

    tree = ast.parse(data)
    new = data

    # funcs = [t for t in tree.body if isinstance(t, ast.FunctionDef)]
    funcs = NodeVisitor()
    funcs.visit(tree)
    funcs = funcs.items
    for i, (func, meta) in enumerate(funcs[:]):
        # print(i, "==", func.name, "==")
        try:
            e0 = func.body[0]
            if not isinstance(e0, ast.Expr):
                continue
            # e0.value is _likely_ a Constant node.
            docstring = e0.value.s
            func_name = func.name
        except AttributeError:
            continue
        if not isinstance(docstring, str):
            continue
        start, nindent, stop = (
            func.body[0].lineno,
            func.body[0].col_offset,
            func.body[0].end_lineno,
        )
        # if not docstring in data:
        #    print(f"skip {file}: {func.name}, can't do replacement yet")
        try:
            new_doc, d_ = compute_new_doc(
                docstring,
                filename,
                level=nindent,
                compact=compact,
                meta=meta,
                func_name=func_name,
            )
            if not unsafe:
                _, d2 = compute_new_doc(
                    docstring,
                    filename,
                    level=nindent,
                    compact=compact,
                    meta=meta,
                    func_name=func_name,
                )
                if not d2._parsed_data == d_._parsed_data:
                    secs1 = {
                        k: v
                        for k, v in d2._parsed_data.items()
                        if v != d_._parsed_data[k]
                    }
                    secs2 = {
                        k: v
                        for k, v in d_._parsed_data.items()
                        if v != d2._parsed_data[k]
                    }
                    raise ValueError(
                        "Numpydoc parsing seem to differ after reformatting, this may be a reformatting bug. Rerun with `velin --unsafe "
                        + str(filename)
                        + "`\n"
                        + str(secs1)
                        + "\n"
                        + str(secs2),
                    )
        except Exception as e:
            print(f"somethign went wrong with {filename}:{docstring}")
            if fail:
                raise
            continue
        if not docstring.strip():
            print("DOCSTRING IS EMPTY !!!", func.name)
        # test(docstring, file)
        if new_doc.strip() and new_doc != docstring:
            # need_changes.append(str(filename) + f":{start}:{func.name}")
            if ('"""' in new_doc) or ("'''" in new_doc):
                print(
                    "SKIPPING", filename, func.name, "triple quote not handled", new_doc
                )
            else:
                # if docstring not in new:
                #    print("ESCAPE issue:", docstring)
                new = new.replace(docstring, new_doc)
    return new


def main():
    config = ConfigParser()
    patterns = []
    if Path("setup.cfg").exists():
        config.read("setup.cfg")
        patterns = config.get("velin", "ignore_patterns", fallback="").split(",")

    parser = argparse.ArgumentParser(description="reformat the docstrigns of some file")
    parser.add_argument(
        "paths",
        metavar="path",
        type=str,
        nargs="+",
        help="Files or folder to reformat",
    )
    parser.add_argument(
        "--context",
        metavar="context",
        type=int,
        default=3,
        help="Number of context lines in the diff",
    )
    parser.add_argument(
        "--unsafe",
        action="store_true",
        help="Lift some safety feature (don't fail if updating the docstring is not indempotent",
    )
    parser.add_argument(
        "--check",
        action="store_true",
        help="Print the list of files/lines number and exit with a non-0 exit status, Use it for CI.",
    )
    parser.add_argument(
        "--no-diff",
        action="store_false",
        dest="print_diff",
        help="Do not print the diff",
    )
    parser.add_argument(
        "--black",
        action="store_true",
        dest="run_black",
        help="Do not run black on examples",
    )
    parser.add_argument("--no-color", action="store_false", dest="do_highlight")
    parser.add_argument("--compact", action="store_true", help="Please ignore")
    parser.add_argument("--no-fail", action="store_false", dest="fail")
    parser.add_argument(
        "--write",
        dest="write",
        action="store_true",
        help="Try to write the updated docstring to the files",
    )

    args = parser.parse_args()
    global BLACK_REFORMAT
    if args.run_black:
        BLACK_REFORMAT = True
    else:
        BLACK_REFORMAT = False
    to_format = []

    for f in args.paths:
        p = Path(f)
        if p.is_dir():
            for sf in p.glob("**/*.py"):
                to_format.append(sf)
        else:
            to_format.append(p)

    def to_skip(file, patterns):
        for p in patterns:
            if re.match(".+" + p + ".+", file):
                return True
        return False

    # need_changes = []
    for file in to_format:
        if to_skip(str(file), patterns):
            print("ignoring", file)
            continue

        try:
            with open(file) as f:
                data = f.read()
        except Exception as e:
            # continue
            continue
            raise RuntimeError(f"Fail reading {file}") from e
        new = reformat_file(data, file, args.compact, args.unsafe, fail=args.fail)
        # test(docstring, file)
        if new != data:

            dold = data.splitlines()
            dnew = new.splitlines()
            diffs = list(
                difflib.unified_diff(
                    dold, dnew, n=args.context, fromfile=str(file), tofile=str(file)
                ),
            )

            if args.print_diff and not args.write:
                code = "\n".join(diffs)

                if args.do_highlight:
                    from pygments import highlight
                    from pygments.formatters import TerminalFormatter
                    from pygments.lexers import DiffLexer

                    code = highlight(code, DiffLexer(), TerminalFormatter())

                print(code)
            if args.write:
                with open(file, "w") as f:
                    f.write(new)
    if args.check:
        if need_changes:
            sys.exit(
                "Some files/functions need updates:\n - " + "\n - ".join(need_changes)
            )
        else:
            sys.exit(0)



from collections import namedtuple
from itertools import chain, cycle

import black


def reformat(lines, indent=4):
    text = "\n".join(lines)
    if "doctest:" in text:
        return lines
    try:
        mode = black.FileMode()
        mode.line_length -= indent + 4
        return black.format_str(text, mode=black.FileMode()).splitlines()
    except Exception as e:
        raise ValueError("could not reformat:" + repr(text)) from e


def insert_promt(lines):
    new = []
    for p, l in zip(chain([">>> "], cycle(["... "])), lines):
        new.append(p + l)
    return new


def splitblank(list):
    items = []
    current = []
    for l in list:
        if not l.strip():
            if current:
                items.append(current)
            current = []
        else:
            current.append(l)
    if current:
        items.append(current)
    return items


InOut = namedtuple("InOut", ["in_", "out"])
Text = namedtuple("Text", ["in_", "out"])


def InOutText(a, b):
    if not a:
        return Text(a, b)
    else:
        return InOut(a, b)


def splitcode(lines):
    """
    Split a block of lines without blank lines into categories.

    Code lines start with >>> or ...,
    then outputs, start with none of the two above.

    """
    items = []
    in_ = []
    out = []
    if not lines[0].startswith(">>>"):
        return [InOutText([], lines)]

    state = "notcode"
    for i, l in enumerate(lines):
        if l.startswith(">>> ") and state == "notcode":
            state = "code"
            if in_ or out:
                items.append(InOutText(in_, out))
            in_, out = [], []

            in_.append(l[4:])
        # ... can appear in pandas output.
        elif (l.startswith("... ") or l.startswith(">>> ")) and state == "code":
            in_.append(l[4:])
        else:
            state = "notcode"
            out.append(l)
    if in_ or out:
        items.append(InOutText(in_, out))
    return items


def reformat_example_lines(ex, indent=4):
    from there import print

    oo = []
    # print(ex)
    try:
        blocks = splitblank(ex)
        for block in blocks:
            # print(block)
            codes = splitcode(block)
            for (in_, out) in codes:
                oo.extend(insert_promt(reformat(in_, indent=4)))
                if out:
                    oo.extend(out)
            oo.append("")
        return oo[:-1]
    except Exception:
        print(block)
        import sys

        raise



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['velin']

package_data = \
{'': ['*']}

install_requires = \
['numpydoc', 'pygments', 'black', 'there']

entry_points = \
{'console_scripts': ['velin = velin:main']}

setup(name='velin',
      version='0.0.6',
      description='French for Vellum',
      author='Matthias Bussonnier',
      author_email='bussonniermatthias@gmail.com',
      url='https://github.com/Carreau/minirst',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.8',
     )
